template="""
from app import cmd, app
import click

'''
    boa cmd %(app)s__create abc
'''
@cmd.command("%(app)s__create")
@click.argument('arg')
def %(app)s__create(arg):
    app.logger.info(arg)

"""


from flask import jsonify, request, make_response, current_app
from flask_admin.base import BaseView
from .urls import expose
import requests as http
import os
import re
import json
from urllib.parse import urljoin
import zlib
import logging


logger = logging.getLogger(__name__)

'''
    admin.add_view(MyView(name='My View', menu_icon_type='glyph', menu_icon_value='glyphicon-home'))

'''

STATIC_PATH = current_app.config["STATIC_PATH"]


to_dict = lambda obj: {k: v for k, v in (obj or {}).items()}


class FacadeView(BaseView):
    '''
        setup = {
            host: microservice return,
            endpoint: {
                frontend: path module or FacadeView.endpoint,
                api: path api,
                js: path js,
                css: path js,
                ctx: path ctx,
            },
            name: name view,
            ico: ico class
        }
    '''

    def __init__(self, setup, *args, **kargs):
        self.setup = setup
        self.template = kargs.pop("template", "admin/template_ctx.html")
        BaseView.__init__(self, *args, **kargs)
    
    def current_user(self):
        """
        docstring
        """
        raise NotImplementedError
    
    def user_dict(self):
        """
        docstring
        """
        raise NotImplementedError
    
    def is_accessible(self):
        raise NotImplementedError

    def _get_dependences(self, type="js"):
        module = re.sub(r"[^\d\w]", "", self.setup["endpoint"]["frontend"].replace("/", "_"))
        path_js = self.setup["endpoint"].get(type, "")
        if not path_js:
            return []
        url = urljoin(self.setup["host"], path_js)
        res_api = http.get(url)
        array_js = res_api.json()
        path_module = os.path.join(STATIC_PATH, os.path.join("dist", module))
        path_module_relative = os.path.join("/static/dist/", module)
        path_manifest = os.path.join(path_module, "manifest.json")
        
        if not os.path.exists(path_module):
            print(":::::USER::", os.getenv("USER"))
            os.mkdir(path_module)

        if os.path.exists(path_manifest):
            manifest = json.load(open(path_manifest, "r"))
        else:
            manifest = {}

        array_js_final = []
        manifest_update = False
        for js in array_js:
            url_js = urljoin(self.setup["host"], js)
            res = http.head(url_js)
            basename = os.path.basename(js)
            path_js = os.path.join(path_module, basename)
            if manifest.get(basename, "") != res.headers["ETag"]:
                manifest_update = True
                manifest[basename] = res.headers["ETag"]
                with open(path_js, "wb+") as f:
                    url_js_tar = urljoin(self.setup["host"], f"z/{js}")
                    res_targz = http.get(url_js_tar)
                    f.write(zlib.decompress(res_targz.content))
                
            array_js_final += ["{}/{}?v={}".format(path_module_relative, basename, str(manifest[basename]))]
        
        if manifest_update:
            json.dump(manifest, open(path_manifest, "w"))

        return array_js_final

    @property
    def extra_js(self):
        return self._get_dependences("js")
    
    @property
    def extra_css(self):
        return self._get_dependences("css")

    @expose('/<path:url>', methods=["GET", "POST", "PUT", "DELETE", "HEAD"])
    def mirror_api(self, url):
        user = self.current_user()
        path_api = self.setup["endpoint"]["api"]
        host = urljoin(self.setup["host"], path_api)
        # logger.info("ms.api: %s" % url)
        url = urljoin(host, url)
        # logger.info("ms.api.join: %s" % url)
        # logger.info("ms.api.method: %s" % request.method)
        try:
            headers = to_dict(request.headers)

            if request.method.lower() in ["post", "put", "delete"]:
                http_action = getattr(http, request.method.lower(), None)

                try:
                    params_args = to_dict(request.args)
                except:
                    params_args = {}
                try:
                    params_json = to_dict(request.json)
                except:
                    params_json = {}
                
                try:
                    params_files = request.files
                except:
                    params_files = None

                res_api = http_action(
                    url=url,
                    params=params_args,
                    json=params_json,
                    files=params_files,
                    # data=request.form,
                    headers=headers)
            else:
                res_api = http.get(url, params=request.args, headers=headers)
                res_api.content
        
            response = make_response(res_api.content, res_api.status_code)
            for k, v in res_api.headers.items():
                response.headers[k] = v
            return response
        except Exception as e:
            logger.exception(e)
            return jsonify({"success": True, "message": str(e)})
    
    def _get_context(self):
        """
            Obtener el contexto
        """
        path_ctx = self.setup["endpoint"].get("ctx", "")

        url = urljoin(self.setup["host"], path_ctx)
        headers = to_dict(request.headers)
        res = http.get(url, headers=headers)
        if res.status_code == 200:
            return res.json()
        return {}
    
    @expose('/')
    def index(self):
        user = self.current_user()
        res = self.render(self.template, i18n=dict(), ctx=self._get_context())
        res = make_response(res)
        ''' Registramos el user de la sesion para el WS ''' 
        res.set_cookie('user_id', str(user.get_id()))
        return res 
    
    @expose('/setup')
    def get_setup(self):
        return jsonify(self.setup)



import click
import os
import sys
from .base import create_app, create_project, tpls
import re
import subprocess
from flask.cli import with_appcontext, FlaskGroup, cli
import importlib


class DefaultCommandGroup(click.Group):
# class DefaultCommandGroup(FlaskGroup):
    """allow a default command for a group"""
    def main(self, *args, **kwargs):

        try:
            importlib.import_module(os.getenv("FLASK_APP", "app"))
        except ModuleNotFoundError:
            pass
            
        return click.Group.main(self, *args, **kwargs)

    def command(self, *args, **kwargs):
        default_command = kwargs.pop('default_command', False)
        if default_command and not args:
            kwargs['name'] = kwargs.get('name', '<>')
        decorator = super(
            DefaultCommandGroup, self).command(*args, **kwargs)

        if default_command:
            def new_decorator(f):
                cmd = decorator(f)
                self.default_command = cmd.name
                return cmd

            return new_decorator

        return decorator

    def resolve_command(self, ctx, args):
        try:
            # test if the command parses
            return super(
                DefaultCommandGroup, self).resolve_command(ctx, args)
        except click.UsageError:
            # command did not parse, assume it is the default command
            args.insert(0, self.default_command)
            return super(
                DefaultCommandGroup, self).resolve_command(ctx, args)
    

@click.group(cls=DefaultCommandGroup, help="Builer Web Services.", invoke_without_command=True)
@click.pass_context
def boa_cli(ctx):

    if not ctx.invoked_subcommand:
        bin_list()


# # cargar los comandos de flask
# for group_name in cli.list_commands({}):
#     group_cmd = cli.get_command({}, group_name)
#     boa_cli.add_command(group_cmd)


@boa_cli.command(
    default_command=True,
    context_settings={
        "ignore_unknown_options": True,
        "allow_extra_args": True,
        # allow_interspersed_args:False
    },
)
@click.argument('args', nargs=-1)
def default(args):
    """Command run without a command"""
    if args:
        bin_run(args)


@boa_cli.command()
@click.argument(
        "project",
        default=os.getcwd(),
        # help="Direccion del proyecto.",
)
@click.option(
        "--app",
        "app",
        default="main",
        help="Nombre de la app default",
)
def init(project, app="main"):
    create_project(project, app)
    click.echo(click.style('OK!', fg='green'))

def callback(ctx, param, value):
    print(ctx, param, value)
    if not value and param:
        ctx.abort()

@boa_cli.command()
@click.argument(
        "name",
        # prompt='Escribe el nomnbre de la aplicación',
        # help="Nombre de la app",
)
@click.option(
        "--dir",
        default=os.getcwd() + "/app",
        help="Nombre de la app default",
)
def add_app(name, dir,):
    if not name:
        print("name is required")
        return
    create_app(dir, name)
    click.echo(click.style('OK!', fg='green'))


@boa_cli.command(name="list")
def bin_list():
    path_bin = os.path.join(os.getcwd(), "bin")
    if os.path.exists(path_bin):
        regex = r"^cmd_([\w\d\_\-]+)\("
        with open(path_bin, "r") as bin:    
            matches = re.findall(regex, bin.read(), re.MULTILINE)
            for match in matches:
                click.echo(f"- {match}")


def common_options(f):
    regex = r"[\-]+"
    if sys.argv[1] == "run":
        argv = sys.argv[3:]
    else:
        argv = sys.argv[2:]

    options_str = " ".join(argv)
    options = re.split(regex, options_str)
    options = filter(lambda d: d, options)
    for option in options:
        # print("option", option)
        if "=" in option:
            option = option.strip().split("=")
        else:
            option = option.strip().split()
        if len(option) == 1:
            f = click.option(
                f"-{option[0]}", f"--{option[0]}",
                is_flag=True
            )(f)
        else:
            f = click.option(
                f"-{option[0]}",f"--{option[0]}",
            )(f)
    return with_appcontext(f)


@boa_cli.command(
    name="run",
    context_settings={
        "ignore_unknown_options": True,
        "allow_extra_args": True,
        # allow_interspersed_args:False
    },
)
@click.argument(
    "script",
)
@click.argument(
    "extra",
    nargs=-1
)
# @common_options
def bin_run(script, extra, **kargs):
    if sys.argv[1] == "run":
        argv = sys.argv[2:]
    else:
        argv = sys.argv[1:]
    path_bin = os.path.join(os.getcwd(), "bin")
    
    if os.path.exists(path_bin):
        
        regex = fr"^cmd_{re.escape(script)}\("
        with open(path_bin, "r") as bin:
            matches = re.findall(regex, bin.read(), re.MULTILINE)
            if len(matches) > 0:
                subprocess.run(
                    # ["bash", "bin", script] + list(extra),
                    ["bash", "bin"] + argv,
                )
            else:
                click.echo(click.style(f'Command [{script}] no found!', blink=True))
    else:
        click.echo(click.style(f'bin no found!', blink=True))



@boa_cli.command(name="reset-bin")
def bin_reset():
    path_bin = os.path.join(os.getcwd(), "bin")
    
    with open(path_bin, "w") as bin:    
        bin.write(tpls.bin)
    
    click.echo(click.style(f'reset!', blink=True))


template="""
from boa.admin.urls import json_render
from app import api


@api.route("/%(app)s")
@json_render
def get_%(app)s_api():
    return {
        "success": True, 
        "message": "Metodo de api para api/%(app)s"
    }

"""


import pytz

def change_timezone(dt, origin='America/Caracas', target='America/Bogota'):
    utc = pytz.timezone(origin)  
    tz = pytz.timezone(target) # timezone you want to convert to from UTC
    try:
        value = utc.localize(dt, is_dst=None).astimezone(pytz.utc)
    except ValueError:
        value = dt 
    local_dt = value.astimezone(tz)
    return local_dt


# from app import admin
import importlib
import logging
from pathlib import Path

logger = logging.getLogger("Boot Admin")


def bootstrap(app, admin=None, types=["config", "api"]):    
    # Register admin of apps
    for type_ in types:
        for package in app.config["INSTALLED_APPS"]:
            if type_ in ["config"]:
                assert not admin is None 
                for file in ["config"]:
                    package_admin = "%s.%s" % (package, file)

                    path = Path(package_admin.replace('.', '/') + '.py')
                    if path.is_file():
                        logger.debug("import %s" % package_admin)
                        module = importlib.import_module(package_admin)
                        register_frontend_views = getattr(module, "register_frontend_views", lambda a: 0)
                        register_api_views = getattr(module, "register_api_views", lambda a: 0)
                        register_frontend_views(admin)
                        register_api_views(admin)
                        logger.debug("imported %s" % package_admin)

                
                # admin.bootstrap_frontends()

            elif type_ in ["api", "routes", "rpc"]:
                for file in ["api", "routes", "rpc"]:
                    package_admin = "%s.%s" % (package, file)
                    path = Path(package_admin.replace('.', '/') + '.py')
                    if path.is_file():
                        logger.debug("api.import %s" % package_admin)
                        __import__(package_admin)
                        logger.debug("api.imported %s" % package_admin)
            else:
                for file in [type_]:
                    package_admin = "%s.%s" % (package, file)
                    path = Path(package_admin.replace('.', '/') + '.py')
                    if path.is_file():
                        __import__(package_admin)
                        logger.debug("import %s" % package_admin)



template = """#!/bin/bash 
#-xe

export DIR_VENV=$PWD/_venv
export FLASK_APP=entrypoint.py

if [ -d "$DIR_VENV" ]; then
  source $DIR_VENV/bin/activate
fi

cmd_install() {
  # cmd_install. Tambien se puede usar para actualizar & agregadas librerias
  if [ ! -d "$DIR_VENV" ]; then
    # sudo apt-get install python3-venv -y
    python3 -m venv $DIR_VENV
    source $DIR_VENV/bin/activate
  fi
  
  if [ -f "requirements-lock.txt" ]; then
    pip install -r requirements-lock.txt
  fi

  pip install -r requirements.txt
  pip freeze -r requirements.txt > requirements-lock.txt
}

cmd_add() {
  # agregar paquetes
  pip install $@
  pip freeze -r requirements.txt > requirements-lock.txt
}

cmd_add-mg() {
  # agregar paquetes & buscarlo primero en el pypi maujagroup 
  pip install $@ -i http://pypi.maujagroup.com:8080/simple\
    --trusted-host pypi.maujagroup.com\
    --extra-index-url https://pypi.org/simple

  pip freeze -r requirements.txt > requirements-lock.txt
}

cmd_lock() {
  # crear lock
  pip freeze -r requirements.txt > requirements-lock.txt
}

cmd_del() {
  # agregar paquetes
  pip uninstall $@
  pip freeze -r requirements.txt > requirements-lock.txt
}


cmd_web() { 
  export ENVFILE=.env
  export FLASK_ENV=development
  flask run
}

cmd_webdev() { 
  export ENVFILE=.env.dev
  flask run
}

cmd_flask() { 
  flask "$@" 
}

cmd_cmd() { 
  flask cmd "$@" 
}


cmd_db() { 
  flask db "$@" 
}

cmd_pip() { 
  pip "$@" 
}

cmd_pos-commit() { 
  echo "pos-commit";
}

cmd_version() {
  # python --version
  flask --version
  pip --version
}

cd "$(dirname "$0")"; _cmd="${1?"cmd is required"}"; shift; "cmd_${_cmd}" "$@"
"""




from app import db
from datetime import datetime


class ExampleModel(db.Model):
    __tablename__ = 'ms_example'

    id= db.Column(db.Integer, primary_key=True, autoincrement=True)
    name=db.Column(db.String(50), nullable=False)
    created_at=db.Column(db.DateTime(), default=datetime.utcnow)
    updated_at=db.Column(db.DateTime(), default=datetime.utcnow)




from flask import current_app, send_file, abort
from flask_admin import expose, AdminIndexView
import flask_admin
import logging
import requests as http
from .urls import *
from .view import *
# import importlib
import tarfile
import os
import io
import inspect
import zlib
from .view import BaseFrontendView, BaseView


logger = logging.getLogger("Boot Admin")


# Create customized index view class that handles login & registration
class IndexView(AdminIndexView):
    def __init__(self, ctx, *args, **kwargs):
        self.ctx = ctx
        AdminIndexView.__init__(self, *args, **kwargs)

    @expose_json('/')
    def index(self):
        # app = importlib.import_module("app").app
        return {
            "success": True,
            "message": current_app.config["NAME"]
        }
    
    @expose('/dev')
    def get_dev(self):
        return """<pre>Jose Angel Delgado.
       _           _____  ___
      (_)___ _____/ /__ \<  /
     / / __ `/ __  /__/ // / 
    / / /_/ / /_/ // __// /  
 __/ /\__,_/\__,_//____/_/   
/___/                        </pre>      
<style>body{ text-align: center; margin-top: 20% }</style>   
"""

    @expose('/tar/static/dist/<path:url>')
    def get_tar(self, url):
        filename = f"static/dist/{url}"
        buffer = io.BytesIO()
        with tarfile.open(fileobj=buffer, mode="w:gz") as tar:
            tar.add(filename, arcname=os.path.basename(filename))
        buffer.seek(0)
        return send_file(buffer,  mimetype="application/tar+gzip")
    
    @expose('/z/static/dist/<path:url>')
    def get_zlib(self, url):
        filename = f"static/dist/{url}"
        if os.path.exists(filename):
            with open(filename, "rb") as fb:
                return zlib.compress(fb.read(), level=9)
        else:
            return abort(404)

    @expose('/wiki.md')
    def get_wiki(self):
        return """Metodo reservado para responder la wiki. de los servicios """
    
    @expose('/reload')
    def get_reload(self):
        try:
            app = current_app
            # print(self.ctx["frontviews"])
            res = http.post(app.config["REGISTER"], json=self.ctx["frontviews"], verify=False)
            logger.debug(res.content)
            return res.content
        except Exception as e:
            logger.exception(e)
            return "No se pudo registrar"
    
    @expose_json('/__egg__')
    @expose_json('/🥚')
    def get_egg(self):
        return self.ctx.get("frontviews", {})
        
    # @expose_json('/config')
    # def get_config(self):
    #     app = current_app
    #     return app.config
        

class Admin(flask_admin.Admin):
    def __init__(self, ctx={}, *args, **kwargs):
        self.ctx = ctx
        flask_admin.Admin.__init__(self, *args, **kwargs)

    def bootstrap_frontends(self, app=None):
        try:
            if not app:
                app = current_app
            
            logger.debug("register: %s" % app.config.get("REGISTER"))
            logger.debug("ctx: %s" % self.ctx)
            for setup in self.ctx.get("frontviews", []):
                setup["host"] = app.config["MS_HOST"]
            if self.ctx.get("frontviews") and app.config.get("REGISTER"):
                res = http.post(app.config["REGISTER"], json=self.ctx["frontviews"], verify=False)
                logger.debug(res.content)
                return res.content

        except Exception as e:
            logger.exception(e)
            return "No se pudo registrar"

    def register_frontend(self, view, endpoint=""):
        # app = self.app
        try:

            frontviews = self.ctx.get("frontviews", [])
            if inspect.isclass(view):
                view = view(endpoint=endpoint, name=endpoint)

            endpoint = view.endpoint or endpoint or endpoint.__class__.__name__.lower()
            
            if endpoint.endswith("/"):
                endpoint = endpoint[:-1]
            
            logger.debug("endpoint: %s" % endpoint)
            self.add_view(view)
            setup = {
                # "host": app.config["MS_HOST"],
                "name": view.name,
                "ico": view.ico,
                "kwargs": getattr(view, "kwargs", {}),
                "endpoint": {
                    "frontend": view.endpoint_frontend,
                    "api": view.endpoint_api or '/',
                    "js": f"{endpoint}/js",
                    "css": f"{endpoint}/css",
                    "ctx": f"{endpoint}/ctx",

                }
            }
            frontviews += [setup]
            self.ctx["frontviews"] = frontviews 
            # print(self.ctx["frontviews"])
        except Exception as e:
            logger.exception(e)

    def register_api(self, view, endpoint="", *args, **kargs):
        endpoint = endpoint or endpoint.__class__.__name__.lower()
        if inspect.isclass(view):
            view = view(endpoint=endpoint, *args, **kargs)
        self.add_view(view)

    def expose(self, url):
        def decorator(klass):
            
            if issubclass(klass, BaseFrontendView):
                self.register_frontend(klass, url)
            elif issubclass(klass, BaseView):
                self.register_api(klass, url)
                 
            return klass
        return decorator 


def create_admin(app=None, **kwargs):
    # Create admin
    ctx = {}
    return Admin(ctx, app, 'MS',
        url="/",
        endpoint="index",
        index_view=IndexView(ctx, url="/", endpoint="index"),
        **kwargs
    )



import json
import gzip
from flask import make_response, jsonify, Response
from functools import wraps
from flask_admin import expose
from .json import JSONEncoderCurstom
from flask import request


def json_render(f):
    @wraps(f)
    def decorated_view(*args, **kwargs):
        body = f(*args, **kwargs)
        try:
            if not isinstance(body, Response):
                if "gzip" in request.headers.get("Accept-Encoding", "") and not "json-not-gzip" in request.args:
                    content = gzip.compress(json.dumps(
                        body, cls=JSONEncoderCurstom).encode('utf8'), 5)
                    response = make_response(content)
                    response.headers['Content-length'] = len(content)
                    response.headers['Content-Encoding'] = 'gzip'
                    response.headers['Content-Type'] = 'application/json'
                    return response
                else:
                    return jsonify(body)
            return body
        except Exception as e:
            return body

    return decorated_view


def expose_json(url='/', methods=('GET',)):
    """
        Use este decorador para responder json
    """
    def wrap(f):
        if not hasattr(f, '_urls'):
            f._urls = []
        f._urls.append((url, methods))

        return json_render(f)
    return wrap


__all__ = [
    "expose",
    "json_render",
    "expose_json" ,  
]



from flask_babel import format_datetime as format_datetime_base, format_decimal as format_decimal_base
import os


FORMAT_DT = os.getenv("FORMAT_DT", "dd/MM/yy HH:mm")


def format_decimal(number, *a, **k):
    try:
        return format_decimal_base(number, *a, **k)
    except:
        return number


def format_datetime(dt, format=FORMAT_DT, **k):
    try:
        return format_datetime_base(dt, format, **k)
    except:
        return ""



import datetime
import decimal
from typing import Generator
from flask import json


class JSONEncoderCurstom(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)

        if isinstance(o, (datetime.datetime, datetime.date)):
            return o.isoformat()
        
        if isinstance(o, Generator):
            o = list(o)

        try:
            return super().default(o)
        except:
            return str(o)



template = """
# Micro - Servicio

Micro-servicio auto generado.

# Instalación
```bash
apt install python-venv
pip install -g bws-boa

boa init .
boa install

cat .env > .env.dev

# configura la db dentro de .env
boa db init

# para preparar la migración
boa db migrate -m "Initial database"

# ejecutar la migración
flask db upgrade

```

# Desarrollo

```bash
boa webdev
```
"""



template = """# generado automaticamente

import os
import sys
from dotenv import load_dotenv

envfile = "../{}".format(os.getenv("ENVFILE", ".env"))

dir_path = os.path.dirname(os.path.realpath(__file__))
load_dotenv(dotenv_path=os.path.join(dir_path, envfile), verbose=True, override=True)

def parse_url_db(engine, user, password, host, port, name):
    return f"{engine}://{user}:{password}@{host}:{port or 3306}/{name}?charset=UTF8MB4&autocommit=false"

TESTING = False
JWT_SECRET_KEY = os.getenv("JWT_SECRET_KEY", 'development')

NAME     = os.getenv("NAME", os.getenv("HOSTNAME"))
HOST     = os.getenv("HOST", "localhost")
PORT     = os.getenv("PORT", 5000)
REGISTER = os.getenv("REGISTER", "http://ip_node")
MS_HOST = os.getenv("MS_HOST")
DEBUG = os.getenv("DEBUG", False)

DB_ENGINE = os.getenv("DB_ENGINE", "mysql+pymysql")
DB_NAME = os.getenv("DB_NAME")
DB_USER = os.getenv("DB_USER")
DB_PASSWORD = os.getenv("DB_PASSWORD")
DB_HOST = os.getenv("DB_HOST")
DB_PORT = os.getenv("DB_PORT", 3306)

SQLALCHEMY_DATABASE_URI = os.getenv("SQLALCHEMY_DATABASE_URI", parse_url_db(
    engine=DB_ENGINE,
    name=DB_NAME,
    user=DB_USER,
    password=DB_PASSWORD,
    port=DB_PORT,
    host=DB_HOST
))
SQLALCHEMY_TRACK_MODIFICATIONS = False
SQLALCHEMY_ENGINE_OPTIONS = {
    "pool_recycle": 490,
    "pool_pre_ping": 60,
}

INSTALLED_APPS = [
    "app.%(app)s",
]
"""
